from pathlib import Path

import pytest
from pkglts.hash_management import compute_hash, modified_file_hash, pth_as_key


@pytest.fixture()
def tmp_pth():
    pth = Path('toto.txt')

    yield pth

    if pth.exists():
        pth.unlink()


def test_pth_as_key_produce_unique_path():
    pth1 = pth_as_key(Path("./toto/titi/"))
    pth2 = pth_as_key(Path("toto/titi"))
    pth3 = pth_as_key(Path("./toto/../toto/titi"))

    assert pth1 == pth2
    assert pth1 == pth3


def test_compute_hash():
    assert compute_hash("toto") == compute_hash("toto")
    assert compute_hash("toto") != compute_hash("titi")


def test_modified_file_raise_error_if_file_not_in_hashmap(tmp_pth):
    with open(tmp_pth, 'w') as f:
        f.write("lorem ipsum\n" * 10)

    with pytest.raises(IOError):
        modified_file_hash(tmp_pth, {})


def test_modified_file_detect_modifications_only_in_preserved_sections(tmp_pth):
    with open(tmp_pth, 'w') as f:
        f.write("lorem ipsum\n" * 10)

    assert not modified_file_hash(tmp_pth, {pth_as_key(tmp_pth): []})


def test_modified_file_if_not_same_preserved_sections(tmp_pth):
    txt = "lorem ipsum\n" * 10

    with open(tmp_pth, 'w') as f:
        f.write("{# pkglts, toto\n")
        f.write(txt)
        f.write("#}\n")

    assert modified_file_hash(tmp_pth, {pth_as_key(tmp_pth): dict(titi="azerty")})


def test_modified_file_detect_modifications_in_preserved_sections(tmp_pth):
    txt = "lorem ipsum\n" * 10
    hv = compute_hash(txt)

    with open(tmp_pth, 'w') as f:
        f.write("{# pkglts, toto\n")
        f.write(txt)
        f.write("#}\n")

    assert not modified_file_hash(tmp_pth, {pth_as_key(tmp_pth): dict(toto=hv)})

    with open(tmp_pth, 'w') as f:
        f.write("{# pkglts, toto\n")
        f.write(txt * 2)
        f.write("#}\n")

    assert modified_file_hash(tmp_pth, {pth_as_key(tmp_pth): dict(toto=hv)})



import pytest
from pkglts.config_management import Config
from pkglts.option.readthedocs.option import OptionReadthedocs


@pytest.fixture()
def opt():
    return OptionReadthedocs('readthedocs')


@pytest.fixture()
def cfg():
    return Config(dict(github={}))


def test_update_parameters(opt, cfg):
    opt.update_parameters(cfg)
    assert len(cfg['readthedocs']) == 1


def test_config_check_project_exists(opt, cfg):
    cfg['readthedocs'] = {'project': ""}
    assert 'readthedocs.project' in opt.check(cfg)


def test_require_option(opt, cfg):
    assert len(tuple(opt.require_option(cfg))) == 3


def test_require(opt, cfg):
    opt.update_parameters(cfg)

    assert len(tuple(opt.require(cfg))) == 0



"""
helper function for formatting documentation.
"""


def fmt_badge(badge, url, txt, fmt):
    """Produce valid img hyperlink.

    Args:
        badge (str): url of img
        url (str): url target of link
        txt (str): associated text
        fmt (str): doc format either 'rst' or 'md'

    Returns:
        (str)
    """
    if not url.startswith("http"):
        url = f"https://{url}"
    if not badge.startswith("http"):
        badge = f"https://{badge}"

    if fmt == 'rst':
        lines = [f".. image:: {badge}",
                 f"    :alt: {txt}",
                 f"    :target: {url}"]
        return "\n" + "\n".join(lines)

    if fmt == 'md':
        return f"[![{txt}]({badge})]({url})"

    raise UserWarning(f"Unknown format '{fmt}'")



import json
from pathlib import Path

import pytest
from pkglts.config import pkg_cfg_file, pkglts_dir
from pkglts.config_management import Config, write_pkg_config
from pkglts.hash_management import get_pkg_hash
from pkglts.manage import init_pkg, regenerate_package, reset_package
from pkglts.small_tools import ensure_created, rmdir


@pytest.fixture()
def tmp_pths():
    pth = Path('toto_mg_rg')
    ensure_created(pth)
    init_pkg(pth)
    with open((pth / pkglts_dir / pkg_cfg_file), 'r') as fhr:
        pkg_cfg = json.load(fhr)

    pkg_cfg['base'] = dict(pkgname='toto',
                           namespace=None,
                           authors=[('moi', 'moi@email.com')],
                           url=None)
    pkg_cfg['src'] = dict(namespace_method="pkg_util")

    cfg = Config(pkg_cfg)
    write_pkg_config(cfg, pth)
    regenerate_package(cfg, pth)

    init_file = pth / "src/toto/__init__.py"

    yield pth, init_file

    rmdir(pth)


def test_reset_removes_files(tmp_pths):
    tmp_dir, init_file = tmp_pths
    with open(tmp_dir / pkglts_dir / pkg_cfg_file, 'r') as fhr:
        pkg_cfg = json.load(fhr)

    tpl_pths = [Path(pth) for pth in get_pkg_hash(tmp_dir)]

    for pth in tpl_pths:
        assert pth.exists()

    cfg = Config(pkg_cfg)
    reset_package(cfg, tmp_dir)

    for pth in tpl_pths:
        assert not pth.exists()


def test_reset_does_not_touch_non_templated_files(tmp_pths):
    tmp_dir, init_file = tmp_pths
    with open(tmp_dir / pkglts_dir / pkg_cfg_file, 'r') as fhr:
        pkg_cfg = json.load(fhr)

    non_tpl_file = init_file.parent / "toto.py"
    non_tpl_file.write_text("# Toto was here")

    assert non_tpl_file.exists()

    cfg = Config(pkg_cfg)
    reset_package(cfg, tmp_dir)

    assert non_tpl_file.exists()



from pathlib import Path

from pkglts.dependency import Dependency
from pkglts.option_object import Option
from pkglts.option_tools import available_options
from pkglts.version import __version__

from . import find_requirements


class OptionReqs(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def update_parameters(self, cfg):
        sec = dict(
            require=[]
        )
        cfg[self._name] = sec

    def check(self, cfg):
        invalids = []

        req = cfg[self._name]['require']

        valid_methods = (None, "pip", "conda", "git")
        if any(dep.get('pkg_mng') not in valid_methods for dep in req):
            invalids.append("reqs.require")

        return invalids

    def require_option(self, cfg):
        return ['base']

    def tools(self, cfg):
        del cfg
        yield find_requirements.parser_find_reqs

    def environment_extensions(self, cfg):
        reqs = requirements(cfg)

        def all_intents():
            intents = {'install', 'doc', 'dvlpt', 'test'}
            for r in reqs:
                intents.update(r.intents)

            return sorted(intents)

        def req(intent=None):
            """For internal use only."""
            if intent is None:
                return reqs

            return [r for r in reqs if intent in r.intents]

        def conda_reqs(intents):
            return fmt_conda_reqs(reqs, intents)

        def pip_reqs(intents):
            return fmt_pip_reqs(reqs, intents)

        cfg.add_test('is_pip_dep', Dependency.is_pip)

        return {
            "intents": all_intents,
            "requirements": req,
            "conda_reqs": conda_reqs,
            "pip_reqs": pip_reqs,
        }


def requirements(cfg):
    """Check all requirements for installed options.

    Args:
        cfg (Config):  current package configuration

    Returns:
        (list): list of required packages
    """
    reqs = {}
    for name in cfg.installed_options():
        try:
            opt = available_options[name]
            for dep in opt.require(cfg):
                reqs[dep.name] = dep
        except KeyError:
            raise KeyError(f"option '{name}' does not exists")

    for dep_def in cfg['reqs']['require']:
        dep = Dependency(**dep_def)
        reqs[dep.name] = dep

    return [reqs[name] for name in sorted(reqs)]


def fmt_conda_reqs(reqs, intents):
    """Produce conda cmd line to install list of requirements.

    Args:
        reqs (list of Dependency): list of requirements objects
        intents (list of str): list of intents for deps

    Returns:
        (str)
    """
    intents = set(intents)
    reqs = [r for r in reqs if r.is_conda(strict=False) and len(r.intents & intents) > 0]
    if len(reqs) == 0:
        return ""

    cmd = "conda install"
    for channel in set(r.channel for r in reqs) - {None}:
        cmd += f" -c {channel}"

    for name in sorted(r.conda_full_name() for r in reqs):
        if " " in name:
            name = f'"{name}"'
        cmd += f" {name}"

    return cmd


def fmt_pip_reqs(reqs, intents):
    """Produce pip cmd line to install list of requirements.

    Args:
        reqs (list of Dependency): list of requirements objects
        intents (list of str): list of intents for deps

    Returns:
        (str)
    """
    intents = set(intents)
    reqs = [r for r in reqs if r.is_pip(strict=True) and len(r.intents & intents) > 0]
    if len(reqs) == 0:
        return ""

    cmd = "pip install"

    for name in sorted(r.pip_full_name() for r in reqs):
        if " " in name:
            name = f'"{name}"'
        cmd += f" {name}"

    return cmd



from pathlib import Path

from pkglts.option.reqs.option import requirements
from pkglts.option_object import Option
from pkglts.version import __version__


class OptionConda(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def require_option(self, cfg):
        return ['pysetup']

    def environment_extensions(self, cfg):
        channels = set(dep.channel for dep in requirements(cfg)) - {None}

        return {"channels": channels}



import pytest
from pkglts.config_management import Config
from pkglts.option.src.option import OptionSrc


@pytest.fixture()
def opt():
    return OptionSrc('src')


@pytest.fixture()
def cfg():
    return Config()


def test_update_parameters(opt, cfg):
    opt.update_parameters(cfg)
    assert len(cfg['src']) == 1


def test_config_check_pkg_names(opt, cfg):
    cfg['src'] = {'namespace_method': "toto"}
    assert 'src.namespace_method' in opt.check(cfg)


def test_require_option(opt, cfg):
    assert len(tuple(opt.require_option(cfg))) == 1


def test_require(opt, cfg):
    opt.update_parameters(cfg)

    assert len(tuple(opt.require(cfg))) == 0



import pytest
from pkglts.config_management import Config
from pkglts.option.coverage.option import OptionCoverage


@pytest.fixture()
def opt():
    return OptionCoverage('coverage')


@pytest.fixture()
def cfg():
    return Config()


def test_require_option(opt, cfg):
    assert len(tuple(opt.require_option(cfg))) == 1


def test_require(opt, cfg):
    cfg['test'] = {'suite_name': 'pytest'}
    opt.update_parameters(cfg)

    assert len(tuple(opt.require(cfg))) == 2

    cfg['test'] = {'suite_name': 'nose'}
    opt.update_parameters(cfg)
    assert len(tuple(opt.require(cfg))) == 1



from pathlib import Path

from pkglts.option.doc import fmt_badge
from pkglts.option_object import Option
from pkglts.version import __version__


class OptionAppveyor(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def update_parameters(self, cfg):
        sec = dict(
            token=""
        )
        cfg[self._name] = sec

    def require_option(self, cfg):
        return ['doc', 'pysetup', 'github']

    def environment_extensions(self, cfg):
        owner = cfg['github']['owner']
        project = cfg['github']['project'].replace("_", "-")
        token = cfg[self._name]['token']

        url = f"ci.appveyor.com/project/{owner}/{project}/branch/master"
        img = f"ci.appveyor.com/api/projects/status/{token}/branch/master?svg=true"
        badge = fmt_badge(img, url, "Appveyor build status", cfg['doc']['fmt'])

        return {"badge": badge}



import pytest
from pkglts.config_management import Config
from pkglts.option.sphinx.option import OptionSphinx


@pytest.fixture()
def opt():
    return OptionSphinx('sphinx')


@pytest.fixture()
def cfg():
    return Config(dict(doc={'fmt': 'rst'}))


def test_update_parameters(opt, cfg):
    opt.update_parameters(cfg)
    assert len(cfg['sphinx']) == 5


def test_config_checks_doc_fmt(opt, cfg):
    cfg['doc'] = {'fmt': 'md'}
    opt.update_parameters(cfg)
    assert 'doc.fmt' in opt.check(cfg)


def test_config_check_sphinx_theme(opt, cfg):
    opt.update_parameters(cfg)

    for theme in (1, None,):
        cfg['sphinx']['theme'] = theme
        assert 'sphinx.theme' in opt.check(cfg)


def test_config_check_sphinx_gallery(opt, cfg):
    opt.update_parameters(cfg)

    assert 'sphinx.gallery' not in opt.check(cfg)

    for pth in ("a" * 256,):
        cfg['sphinx']['gallery'] = pth
        assert 'sphinx.gallery' in opt.check(cfg)


def test_require_option(opt, cfg):
    assert len(tuple(opt.require_option(cfg))) == 2


def test_require(opt, cfg):
    opt.update_parameters(cfg)

    assert len(tuple(opt.require(cfg))) == 1



import pytest
from pkglts.config_management import Config
from pkglts.option.appveyor.option import OptionAppveyor


@pytest.fixture()
def opt():
    return OptionAppveyor('appveyor')


@pytest.fixture()
def cfg():
    return Config(dict(base={}, travis={}))


def test_update_parameters(opt, cfg):
    opt.update_parameters(cfg)
    assert len(cfg['appveyor']) == 1


def test_config_does_nothing(opt, cfg):
    cfg['appveyor'] = {'token': 'toto'}
    assert len(opt.check(cfg)) == 0


def test_require_option(opt, cfg):
    assert len(tuple(opt.require_option(cfg))) == 3


def test_require(opt, cfg):
    opt.update_parameters(cfg)

    assert len(tuple(opt.require(cfg))) == 0



import pytest
from pkglts.config_management import Config
from pkglts.option.pysetup.option import OptionPysetup


@pytest.fixture()
def opt():
    return OptionPysetup('pysetup')


@pytest.fixture()
def cfg():
    return Config()


def test_update_parameters(opt, cfg):
    opt.update_parameters(cfg)
    assert len(cfg['pysetup']) == 1


def test_config_check_intended_version_exists(opt, cfg):
    cfg['pysetup'] = {'intended_versions': []}
    assert 'pysetup.intended_versions' in opt.check(cfg)


def test_require_option(opt, cfg):
    assert len(tuple(opt.require_option(cfg))) == 5


def test_require(opt, cfg):
    opt.update_parameters(cfg)

    assert len(tuple(opt.require(cfg))) == 0



from pathlib import Path

import pytest
from pkglts.manage import clean
from pkglts.small_tools import ensure_created, rmdir


@pytest.fixture()
def tmp_dir():
    pth = Path('toto_manage_clean')
    ensure_created(pth)

    yield pth

    rmdir(pth)


def test_clean_remove_pyc_files(tmp_dir):
    pth = tmp_dir / "toto.pyc"
    pth.write_text("toto")

    clean(tmp_dir)

    assert not pth.exists()


def test_clean_remove_pycache_directories(tmp_dir):
    pycache = tmp_dir / "__pycache__"
    ensure_created(pycache)
    pth = pycache / "toto.py"
    pth.write_text("toto")

    clean(tmp_dir)

    assert not pycache.exists()
    assert not pth.exists()


def test_clean_do_not_remove_py_files(tmp_dir):
    pth = tmp_dir / "toto.py"
    pth.write_text("toto")

    clean(tmp_dir)

    assert pth.exists()


def test_clean_do_not_remove_hidden_files(tmp_dir):
    pth = tmp_dir / ".toto.pyc"
    pth.write_text("toto")

    clean(tmp_dir)

    assert pth.exists()


def test_clean_do_not_explore_hidden_directories(tmp_dir):
    hidden = tmp_dir / ".test"
    ensure_created(hidden)
    pth = hidden / "toto.py"
    pth.write_text("toto")

    clean(tmp_dir)

    assert hidden.exists()
    assert pth.exists()


def test_clean_do_not_explore_clean_no_directories(tmp_dir):
    hidden = tmp_dir / "test"
    ensure_created(hidden)
    pth = hidden / "toto.py"
    pth.write_text("toto")

    (hidden / "clean.no").touch()

    clean(tmp_dir)

    assert hidden.exists()
    assert pth.exists()


def test_clean_remove_dist_build_directories(tmp_dir):
    for name in ("dist", "build"):
        ensure_created(tmp_dir / name)

    clean(tmp_dir)

    assert len(tuple(tmp_dir.iterdir())) == 0



from pathlib import Path
from random import random

import pytest
from pkglts.config_management import Config, DEFAULT_CFG
from pkglts.hash_management import pth_as_key
from pkglts.manage_tools import find_templates, render_template
from pkglts.small_tools import ensure_created, ensure_path, rmdir


@pytest.fixture()
def tmp_dir():
    pth = Path("takapouet")

    ensure_created(pth)
    ensure_created(pth / "src")
    ensure_created(pth / "tgt")

    yield pth

    rmdir(pth)


def test_find_templates_walk_all_files_in_src_dir(tmp_dir):
    ensure_created(tmp_dir / "src/sub")
    fnames = ('toto.txt', 'titi.txt', 'sub/toto.txt')
    for fname in fnames:
        pth = tmp_dir / "src" / fname
        pth.write_text("lorem ipsum")

    cfg = Config(DEFAULT_CFG)
    rg_tree = {}
    find_templates(tmp_dir / 'src', tmp_dir / 'tgt', cfg, rg_tree)

    for fname in fnames:
        assert pth_as_key(tmp_dir / 'tgt' / fname) in rg_tree


def test_find_templates_renders_path_names(tmp_dir):
    ensure_created(tmp_dir / "src/{{ custom_name }}")
    fnames = ('{{ custom_name }}.txt', 'titi.txt',
              '{{ custom_name }}/{{ custom_name }}.txt')
    for fname in fnames:
        pth = tmp_dir / "src" / fname
        pth.write_text("lorem ipsum")

    cfg = Config(DEFAULT_CFG)
    cfg._env.globals['custom_name'] = 'toto'
    rg_tree = {}
    find_templates(tmp_dir / 'src', tmp_dir / 'tgt', cfg, rg_tree)

    fnames = ('toto.txt', 'titi.txt', 'toto/toto.txt')
    for fname in fnames:
        assert pth_as_key(tmp_dir / 'tgt' / fname) in rg_tree


def test_find_templates_ignores_specific_names(tmp_dir):
    ensure_created(tmp_dir / "src/{{ custom_name }}")
    fnames = ('{{ custom_name }}.txt', 'titi.txt',
              '{{ custom_name }}/{{ custom_name }}.txt')
    for fname in fnames:
        pth = tmp_dir / "src" / fname
        pth.write_text("lorem ipsum")

    cfg = Config(DEFAULT_CFG)
    cfg._env.globals['custom_name'] = '_'
    rg_tree = {}
    find_templates(tmp_dir / 'src', tmp_dir / 'tgt', cfg, rg_tree)

    assert pth_as_key(tmp_dir / 'tgt/titi.txt') in rg_tree
    fnames = ('_.txt', '_/_.txt')
    for fname in fnames:
        assert pth_as_key(tmp_dir / 'tgt' / fname) not in rg_tree


def test_find_templates_handles_src_directory_no_namespace(tmp_dir):
    pth = tmp_dir / "src/src/{{ base.pkgname }}/test.txt"
    ensure_path(pth)
    pth.write_text("lorem ipsum")

    pkg_cfg = dict(DEFAULT_CFG)
    pkg_cfg['base'] = dict(pkgname='toto', namespace=None)
    cfg = Config(pkg_cfg)
    rg_tree = {}
    find_templates(tmp_dir / 'src', tmp_dir / 'tgt', cfg, rg_tree)

    tgt = tmp_dir / "tgt/src/toto/test.txt"
    assert pth_as_key(tgt) in rg_tree


def test_fin_templates_handles_src_directory_with_namespace(tmp_dir):
    pth = tmp_dir / "src/src/{{ base.pkgname }}/test.txt"
    ensure_path(pth)
    pth.write_text("lorem ipsum")

    pkg_cfg = dict(DEFAULT_CFG)
    pkg_cfg['base'] = dict(pkgname='toto', namespace='myns', namespace_method='pkg_utils')
    cfg = Config(pkg_cfg)
    rg_tree = {}
    find_templates(tmp_dir / 'src', tmp_dir / 'tgt', cfg, rg_tree)

    tgt_dir = tmp_dir / "tgt"
    assert pth_as_key(tgt_dir / "src/myns/__init__.py") in rg_tree
    assert pth_as_key(tgt_dir / "src/myns/toto/test.txt") in rg_tree


def test_render_template_renders_file_content(tmp_dir):
    src_pth = tmp_dir / "src/test.txt"
    tgt_pth = tmp_dir / "tgt/test.txt"
    src_pth.write_text("{{ 'lorem ipsum'|upper }}")

    cfg = Config(DEFAULT_CFG)
    render_template([src_pth], tgt_pth, cfg, {})

    assert tgt_pth.read_text() == 'LOREM IPSUM'


def test_render_template_overwrites_unprotected_files(tmp_dir):
    src_pth = tmp_dir / "src/test.txt"
    tgt_pth = tmp_dir / "tgt/test.txt"
    src_pth.write_text("{# pkglts, b0\n{{ random() }}\n#}\n")

    cfg = Config(DEFAULT_CFG)
    cfg._env.globals['random'] = random
    render_template([src_pth], tgt_pth, cfg, {})

    cnt0 = tgt_pth.read_text()

    render_template([src_pth], tgt_pth, cfg, {})

    cnt1 = tgt_pth.read_text()

    assert cnt0 != cnt1


def test_render_template_does_not_overwrite_protected_files(tmp_dir):
    src_pth = tmp_dir / "src/test.txt"
    tgt_pth = tmp_dir / "tgt/test.txt"
    src_pth.write_text("{# pkglts, b0\n{{ random() }}\n#}\n")

    cfg = Config(DEFAULT_CFG)
    cfg._env.globals['random'] = random
    render_template([src_pth], tgt_pth, cfg, {})

    cnt0 = tgt_pth.read_text()

    overwrite = {pth_as_key(tgt_pth): False}
    render_template([src_pth], tgt_pth, cfg, overwrite)

    cnt1 = tgt_pth.read_text()

    assert cnt0 == cnt1


def test_render_template_does_not_overwrite_outside_protected_blocks(tmp_dir):
    src_pth = tmp_dir / "src/test.txt"
    tgt_pth = tmp_dir / "tgt/test.txt"
    src_pth.write_text("{# pkglts, b0\nLOREM IPSUM\n#}\n")

    tgt_pth.write_text("Toto start\n{# pkglts, b0\nWTF?\n#}\nToto end\n")

    cfg = Config(DEFAULT_CFG)
    render_template([src_pth], tgt_pth, cfg, {})

    assert tgt_pth.read_text() == "Toto start\n{# pkglts, b0\nLOREM IPSUM\n#}\nToto end\n"



"""
TODO Useless?
"""


def write_file(pth, content):
    """Write the content of a file on a local path and
    register associated hash for further modification
    tests.

    Args:
        pth (str): path to the new created file
        content (str): content to write on disk

    Returns:
        None
    """
    with open(pth, 'wb') as f:
        f.write(content.encode("utf-8"))



import pytest
from pkglts.config_management import Config
from pkglts.option.gitlab.option import OptionGitlab


@pytest.fixture()
def opt():
    return OptionGitlab('gitlab')


@pytest.fixture()
def cfg():
    return Config()


def test_update_parameters(opt, cfg):
    opt.update_parameters(cfg)
    assert len(cfg['gitlab']) == 4


def test_config_check_project_exists(opt, cfg):
    cfg['gitlab'] = {'owner': "", 'project': "", 'server': "", "url": ""}
    assert 'gitlab.project' in opt.check(cfg)


def test_require_option(opt, cfg):
    assert len(tuple(opt.require_option(cfg))) == 1


def test_require(opt, cfg):
    opt.update_parameters(cfg)

    assert len(tuple(opt.require(cfg))) == 0



from pathlib import Path

from pkglts.dependency import Dependency
from pkglts.local import pkg_full_name
from pkglts.option.doc import fmt_badge
from pkglts.option_object import Option
from pkglts.version import __version__


class OptionPypi(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def update_parameters(self, cfg):
        sec = dict(
            classifiers=[
                "Development Status :: 2 - Pre-Alpha",
                "Intended Audience :: Developers",
                "License :: OSI Approved :: BSD License",
                "Natural Language :: English"
            ],
            servers=[
                dict(name="pypi", url="https://upload.pypi.org/legacy/"),
                dict(name="test", url="https://testpypi.python.org/pypi")
            ]
        )
        cfg[self._name] = sec

    def check(self, cfg):
        invalids = []
        classifiers = cfg[self._name]['classifiers']

        if not classifiers:
            invalids.append("pypi.classifiers")

        return invalids

    def require_option(self, cfg):
        return ['doc', 'pysetup']

    def require(self, cfg):
        del cfg
        yield Dependency('twine', intent='dvlpt')

    def environment_extensions(self, cfg):
        ext = {"auto_classifiers": auto_classifiers(cfg)}

        servers = cfg['pypi']['servers']
        if servers and servers[0]['name'] == 'pypi':
            url = f"badge.fury.io/py/{pkg_full_name(cfg)}"
            img = f"{url}.svg"
            ext['badge'] = fmt_badge(img, url, "PyPI version", cfg['doc']['fmt'])

        return ext


def auto_classifiers(cfg):
    """Generate a list of classifiers for pypi from all sections of config.

    Args:
        cfg (Config):  current package configuration

    Returns:
        (list of str)
    """
    items = set(cfg['pypi']['classifiers'])

    # add license item
    # TODO

    # add intended versions items
    intended_versions = cfg['pysetup']['intended_versions']
    if intended_versions:
        items.add("Programming Language :: Python")

        ver_cla_tpl = "Programming Language :: Python :: {}.{}"
        major_versions = set()
        for ver in intended_versions:
            items.add(ver_cla_tpl.format(ver[0], ver[1]))
            major_versions.add(ver[0])

        ver_cla_tpl = "Programming Language :: Python :: {}"
        for ver in major_versions:
            items.add(ver_cla_tpl.format(ver))

        if len(major_versions) == 1:
            ver, = major_versions
            items.add(f"Programming Language :: Python :: {ver} :: Only")

    return sorted(items)



""" Contains functions to manage the structure of the package.

Use 'setup.py' for common tasks.
"""
import logging
from pathlib import Path
from shutil import rmtree

from .config import pkg_hash_file, pkglts_dir
from .config_management import (Config, DEFAULT_CFG,
                                get_pkg_config, write_pkg_config)
from .hash_management import (get_pkg_hash, modified_file_hash,
                              pth_as_key, write_pkg_hash)
from .manage_tools import (check_option_parameters, find_templates,
                           render_template, update_opt)
from .option_tools import available_options, get_user_permission
from .small_tools import ensure_created

LOGGER = logging.getLogger(__name__)


def init_pkg(rep="."):
    """Initialise a package in given directory.

    Args:
        rep (Path): directory to create pkg into, default current

    Returns:
        None
    """
    loc_pkglts_dir = rep / pkglts_dir
    ensure_created(loc_pkglts_dir)

    LOGGER.info("init package")

    for name in ("regenerate.no", "clean.no"):
        (loc_pkglts_dir / name).touch()

    try:
        cfg = get_pkg_config(rep)
    except FileNotFoundError:
        cfg = Config(DEFAULT_CFG)
    write_pkg_config(cfg, rep)

    if not (loc_pkglts_dir / pkg_hash_file).exists():
        write_pkg_hash({}, rep)


def clean(rep="."):
    """Thorough cleaning of all arborescence rooting at rep.

    Todo: exception list instead of hardcoded one

    Args:
        rep (Path): default ".", top directory to clean

    Returns:
        None
    """
    root_dir = Path(rep)
    try:
        cfg = get_pkg_config(rep)

        try:
            doc_dir = root_dir / cfg['sphinx']['doc_dir']
            for pth in (doc_dir / "build", doc_dir / "_dvlpt"):
                if pth.exists():
                    rmtree(pth)
        except KeyError:
            pass
    except FileNotFoundError:
        pass

    for name in ("build", "dist"):
        pth = root_dir / name
        if pth.exists():
            rmtree(pth)

    for dir_pth in tuple(root_dir.glob("**/")):
        if dir_pth.exists():  # could have been removed with a previous rmtree
            if not dir_pth.name.startswith('.') and len(tuple(dir_pth.glob("clean.no"))) == 0:
                if dir_pth.name == '__pycache__':
                    rmtree(dir_pth)
                else:
                    for pth in dir_pth.glob("*.pyc"):
                        if not pth.name.startswith('.'):
                            pth.unlink()
                    for pth in dir_pth.glob("*.pyo"):
                        if not pth.name.startswith('.'):
                            pth.unlink()


def add_option(name, cfg):
    """Add a new option to this package.

    Notes: See the list of available options online

    Args:
        name (str): name of option to add
        cfg (Config):  current package configuration

    Returns:
        (Config): updated package configuration
    """
    if name in cfg.installed_options():
        raise UserWarning(f"option '{name}' already included in this package")

    return update_opt(name, cfg)


def install_example_files(option, cfg, target=Path(".")):
    """Install example files associated to an option.

    Args:
        option (str): name of option
        cfg (Config):  current package configuration
        target (Path): target directory to write into

    Returns:
        (bool): whether operation succeeded or not
    """
    if option not in cfg.installed_options():
        LOGGER.warning("please install option before example files")
        return False

    opt = available_options[option]
    ex_dir = opt.example_dir()
    if ex_dir is None:
        LOGGER.info("option does not provide any example")
        return False

    rg_tree = {}
    find_templates(ex_dir, target, cfg, rg_tree)
    for tgt_key, src_pths in rg_tree.items():
        render_template(src_pths, Path(tgt_key), cfg, {})

    return True


def _manage_conflicts(hm_ref, overwrite):
    """Check if files have been modified.

    Args:
        hm_ref (dict): reference hash for blocks
        overwrite (bool): whether or not, by default,
                         overwrite user modified files

    Returns:
        (dict of str, bool): file path, overwrite
    """
    conflicted = []
    for tgt_key in hm_ref:
        tgt_pth = Path(tgt_key)
        if tgt_pth.exists() and modified_file_hash(tgt_pth, hm_ref):
            conflicted.append(tgt_pth)
        # else file disappeared, regenerate_dir will reload it if managed by pkglts

    overwrite_file = {}
    if conflicted:
        if overwrite:
            for pth in conflicted:
                LOGGER.debug("conflicted, '%s'", pth)
                overwrite_file[pth_as_key(pth)] = True
        else:
            for pth in conflicted:
                LOGGER.warning("A non editable section of %s has been modified", pth)
                overwrite_file[pth_as_key(pth)] = get_user_permission("overwrite", False)

    return overwrite_file


def regenerate_package(cfg, target=Path("."), overwrite=False):
    """Rebuild all automatically generated files.

    Args:
        cfg (Config):  current package configuration
        target (Path): target directory to write into
        overwrite (bool): default False, whether or not
                         to overwrite user modified files

    Returns:
        None
    """
    # check consistency of env params
    invalids = []
    for option in cfg.installed_options():
        for name in check_option_parameters(option, cfg):
            invalids.append((option, name))

    if invalids:
        for option, param in invalids:
            LOGGER.warning("param %s is not valid for '%s'", param, option)

        return False

    # check for potential conflicts
    hm_ref = get_pkg_hash(target)
    overwrite_file = _manage_conflicts(hm_ref, overwrite)

    # find template files associated with installed options
    rg_tree = {}
    for name in cfg.installed_options(return_sorted=True):
        opt = available_options[name]
        resource_dir = opt.resource_dir()
        if resource_dir is None:
            LOGGER.info("option %s does not provide files", name)
        else:
            LOGGER.info("find template for option %s", name)
            find_templates(resource_dir, target, cfg, rg_tree)

    # render all templates
    hmap = {}
    for tgt_key, src_pths in rg_tree.items():
        loc_map = render_template(src_pths, Path(tgt_key), cfg, overwrite_file)
        if loc_map is not None:  # non binary file
            hmap[tgt_key] = loc_map

    hm_ref.update(hmap)
    write_pkg_hash(hm_ref, target)


def regenerate_option(cfg, name, target=".", overwrite=False):
    """Call the regenerate function of a given option

    Args:
        cfg (Config):  current package configuration
        name: (str) name of option
        target: (Path) target directory to write into
        overwrite (bool): default False, whether or not
                         to overwrite user modified files

    Returns:
        None
    """
    # test existence of option regenerate module
    try:
        opt = available_options[name]
        opt.regenerate(cfg, target, overwrite)
    except KeyError:
        raise KeyError(f"option '{name}' does not exists")


def reset_package(cfg, target=Path(".")):
    """Remove all templated files from packages.

    Args:
        cfg (Config):  current package configuration
        target: (Path) target directory to write into

    Returns:
        None
    """
    # check consistency of env params
    invalids = []
    for option in cfg.installed_options():
        for name in check_option_parameters(option, cfg):
            invalids.append((option, name))

    if invalids:
        for option, param in invalids:
            LOGGER.warning("param %s is not valid for '%s'", param, option)

        return False

    hm_ref = get_pkg_hash(target)

    # find template files associated with installed options
    rg_tree = {}
    for name in cfg.installed_options(return_sorted=True):
        opt = available_options[name]
        resource_dir = opt.resource_dir()
        if resource_dir is None:
            LOGGER.info("option %s does not provide files", name)
        else:
            LOGGER.info("find template for option %s", name)
            find_templates(resource_dir, target, cfg, rg_tree)

    # remove all templates
    for tgt_key, src_pths in rg_tree.items():
        LOGGER.debug("remove file '%s'", tgt_key)
        try:
            Path(tgt_key).unlink()
        except FileNotFoundError:
            LOGGER.debug("File '%s' does not exist", tgt_key)

        del hm_ref[tgt_key]

    write_pkg_hash(hm_ref, target)



"""
Base setting for logging in the package.
This module is executed when importing pkglts.
"""
import logging
import logging.handlers

from .config import pkglts_dir


def rolling_logs(logger):
    """Associate log files using a rolling mechanism"""
    nb_logs = 5
    tpl = pkglts_dir / "info.log.%d"
    pth = pkglts_dir / f"info.{nb_logs:d}.log"
    if pth.exists():
        pth.unlink()

    for i in range(nb_logs, 0, -1):
        pth = pkglts_dir / f"info.{i - 1:d}.log"
        if pth.exists():
            pth.rename(pkglts_dir / f"info.{i:d}.log")

    pth = pkglts_dir / "info.log"
    if pth.exists():
        pth.rename(tpl % 0)

    info_ch = logging.FileHandler(str(pth), 'w')
    logger.addHandler(info_ch)


def main(verbosity):
    """Set logging configuration.

    Args:
        verbosity (int): verbosity level

    Returns:
        None
    """
    if verbosity > 2:
        verbosity = 2
    vlevel = [logging.WARNING, logging.INFO, logging.DEBUG][verbosity]

    fmt = logging.Formatter('%(asctime)s %(levelname)s (%(name)s): %(message)s')

    wng_ch = logging.StreamHandler()
    wng_ch.setLevel(vlevel)
    wng_ch.setFormatter(fmt)

    logger = logging.getLogger("pkglts")
    logger.setLevel(vlevel)
    logger.addHandler(wng_ch)

    # try:
    #     # rolling_logs(logger)
    #     info_ch = logging.StreamHandler()
    #     info_ch.setLevel(logging.INFO)
    #     info_ch.setFormatter(fmt)
    #
    #     logger.addHandler(info_ch)
    # except IOError:
    #     print("no info handler")


if __name__ == '__main__':
    main(0)
